"""Create a new L3 Core Service subscription including GÉANT IP and IAS."""

from typing import Any
from uuid import uuid4

from orchestrator.forms import FormPage
from orchestrator.forms.validators import Label
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, done, step, workflow
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from pydantic import BaseModel, ConfigDict, Field, computed_field
from pydantic_forms.validators import Divider

from gso.products.product_blocks.bgp_session import BGPSession, IPFamily, IPTypes
from gso.products.product_blocks.l3_core_service import AccessPortInactive
from gso.products.product_blocks.service_binding_port import BFDSettings, ServiceBindingPortInactive
from gso.products.product_types.edge_port import EdgePort
from gso.products.product_types.l3_core_service import L3CoreService, L3CoreServiceInactive
from gso.services.lso_client import LSOState, lso_interaction
from gso.services.partners import get_partner_by_id
from gso.utils.helpers import (
    active_edge_port_selector,
    partner_choice,
)
from gso.utils.shared_enums import APType, SBPType
from gso.utils.types.ip_address import IPv4AddressType, IPV4Netmask, IPv6AddressType, IPV6Netmask
from gso.utils.types.tt_number import TTNumber
from gso.utils.types.virtual_identifiers import VLAN_ID


def initial_input_form_generator(product_name: str) -> FormGenerator:
    """Gather input from the operator to build a new subscription object."""

    class CreateL3CoreServiceForm(FormPage):
        model_config = ConfigDict(title=f"{product_name} - Select partner")

        tt_number: TTNumber
        partner: partner_choice()  # type: ignore[valid-type]

    initial_user_input = yield CreateL3CoreServiceForm

    class EdgePortSelection(BaseModel):
        edge_port: active_edge_port_selector(partner_id=initial_user_input.partner)  # type: ignore[valid-type]
        ap_type: APType

    class EdgePortSelectionForm(FormPage):
        model_config = ConfigDict(title=f"{product_name} - Select Edge Ports")
        info_label: Label = Field(
            f"Please select the Edge Ports where this {product_name} service will terminate", exclude=True
        )

        edge_port: EdgePortSelection

    selected_edge_port = yield EdgePortSelectionForm

    class BFDSettingsForm(BaseModel):
        bfd_enabled: bool = False
        bfd_interval_rx: int | None = Field(default=None, examples=["BFD RX defaults"])
        bfd_interval_tx: int | None = None
        bfd_multiplier: int | None = None

    class IPv4BGPPeer(BaseModel):
        peer_address: IPv4AddressType
        authentication_key: str | None = None
        has_custom_policies: bool = False
        bfd_enabled: bool = False
        multipath_enabled: bool = False
        is_passive: bool = False
        add_v4_multicast: bool = Field(default=False, exclude=True)
        send_default_route: bool = False

        @computed_field  # type: ignore[misc]
        @property
        def families(self) -> list[IPFamily]:
            return [IPFamily.V4UNICAST, IPFamily.V4MULTICAST] if self.add_v4_multicast else [IPFamily.V4UNICAST]

        @computed_field  # type: ignore[misc]
        @property
        def ip_type(self) -> IPTypes:
            return IPTypes.IPV4

    class IPv6BGPPeer(BaseModel):
        peer_address: IPv6AddressType
        authentication_key: str | None = None
        has_custom_policies: bool = False
        bfd_enabled: bool = False
        multipath_enabled: bool = False
        is_passive: bool = False
        add_v6_multicast: bool = Field(default=False, exclude=True)
        send_default_route: bool = False

        @computed_field  # type: ignore[misc]
        @property
        def families(self) -> list[IPFamily]:
            return [IPFamily.V6UNICAST, IPFamily.V6MULTICAST] if self.add_v6_multicast else [IPFamily.V6UNICAST]

        @computed_field  # type: ignore[misc]
        @property
        def ip_type(self) -> IPTypes:
            return IPTypes.IPV6

    class BindingPortInputForm(FormPage):
        model_config = ConfigDict(title=f"{product_name} - Configure Edge Port")
        info_label: Label = Field("Please configure the Service Binding Ports for the Edge Port.", exclude=True)
        current_ep_label: Label = Field(
            f"Currently configuring on {EdgePort.from_subscription(selected_edge_port.edge_port.edge_port).description}"
            f" (Access Port type: {selected_edge_port.edge_port.ap_type})",
            exclude=True,
        )

        geant_sid: str
        is_tagged: bool = False
        vlan_id: VLAN_ID
        custom_firewall_filters: bool = False
        divider: Divider = Field(None, exclude=True)
        v4_label: Label = Field("IPV4 SBP interface params", exclude=True)
        ipv4_address: IPv4AddressType
        ipv4_mask: IPV4Netmask
        v4_bfd_settings: BFDSettingsForm = BFDSettingsForm(bfd_enabled=False)
        divider2: Divider = Field(None, exclude=True)
        v6_label: Label = Field("IPV6 SBP interface params", exclude=True)
        ipv6_address: IPv6AddressType
        ipv6_mask: IPV6Netmask
        v6_bfd_settings: BFDSettingsForm = BFDSettingsForm(bfd_enabled=False)
        divider3: Divider = Field(None, exclude=True)
        v4_bgp_peer: IPv4BGPPeer
        v6_bgp_peer: IPv6BGPPeer

    binding_port_input_form = yield BindingPortInputForm
    binding_port_input = binding_port_input_form.model_dump() | {
        "bgp_peers": [
            binding_port_input_form.v4_bgp_peer.model_dump(),
            binding_port_input_form.v6_bgp_peer.model_dump(),
        ]
    }

    return (
        initial_user_input.model_dump()
        | selected_edge_port.model_dump()
        | {"binding_port_input": binding_port_input, "product_name": product_name}
    )


@step("Create subscription")
def create_subscription(product: UUIDstr, partner: str) -> State:
    """Create a new subscription object in the database."""
    subscription = L3CoreServiceInactive.from_product_id(product, partner)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: L3CoreServiceInactive, edge_port: dict, binding_port_input: dict, product_name: str
) -> State:
    """Take all user inputs and use them to populate the subscription model."""
    edge_port_fqdn_list = []
    edge_port_subscription = EdgePort.from_subscription(edge_port["edge_port"])
    sbp_bgp_session_list = [
        BGPSession.new(subscription_id=uuid4(), rtbh_enabled=True, is_multi_hop=True, **session)
        for session in binding_port_input["bgp_peers"]
    ]
    service_binding_port = ServiceBindingPortInactive.new(
        subscription_id=uuid4(),
        v4_bfd_settings=BFDSettings.new(subscription_id=uuid4(), **(binding_port_input.pop("v4_bfd_settings"))),
        v6_bfd_settings=BFDSettings.new(subscription_id=uuid4(), **(binding_port_input.pop("v6_bfd_settings"))),
        **binding_port_input,
        bgp_session_list=sbp_bgp_session_list,
        sbp_type=SBPType.L3,
        edge_port=edge_port_subscription.edge_port,
    )
    subscription.l3_core_service.ap_list.append(
        AccessPortInactive.new(
            subscription_id=uuid4(),
            ap_type=edge_port["ap_type"],
            sbp=service_binding_port,
        )
    )
    edge_port_fqdn_list.append(edge_port_subscription.edge_port.node.router_fqdn)

    subscription.description = f"{product_name} service"
    partner_name = get_partner_by_id(subscription.customer_id).name
    return {"subscription": subscription, "edge_port_fqdn_list": edge_port_fqdn_list, "partner_name": partner_name}


@step("[DRY RUN] Deploy service binding port")
def provision_sbp_dry(
    subscription: dict[str, Any], process_id: UUIDstr, tt_number: str, edge_port_fqdn_list: list[str], partner_name: str
) -> LSOState:
    """Perform a dry run of deploying Service Binding Ports."""
    extra_vars = {
        "subscription": subscription,
        "partner_name": partner_name,
        "dry_run": True,
        "verb": "deploy",
        "object": "sbp",
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Deploy config for {subscription["description"]}",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/l3_core_service.yaml",
        "inventory": {"all": {"hosts": dict.fromkeys(edge_port_fqdn_list)}},
        "extra_vars": extra_vars,
    }


@step("[FOR REAL] Deploy service binding port")
def provision_sbp_real(
    subscription: dict[str, Any], process_id: UUIDstr, tt_number: str, edge_port_fqdn_list: list[str], partner_name: str
) -> LSOState:
    """Deploy Service Binding Ports."""
    extra_vars = {
        "subscription": subscription,
        "partner_name": partner_name,
        "dry_run": False,
        "verb": "deploy",
        "object": "sbp",
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Deploy config for {subscription["description"]}",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/l3_core_service.yaml",
        "inventory": {"all": {"hosts": dict.fromkeys(edge_port_fqdn_list)}},
        "extra_vars": extra_vars,
    }


@step("Check service binding port functionality")
def check_sbp_functionality(subscription: dict[str, Any], edge_port_fqdn_list: list[str]) -> LSOState:
    """Check functionality of deployed Service Binding Ports."""
    extra_vars = {"subscription": subscription, "verb": "check", "object": "sbp"}

    return {
        "playbook_name": "gap_ansible/playbooks/l3_core_service.yaml",
        "inventory": {"all": {"hosts": dict.fromkeys(edge_port_fqdn_list)}},
        "extra_vars": extra_vars,
    }


@step("[DRY RUN] Deploy BGP peers")
def deploy_bgp_peers_dry(
    subscription: dict[str, Any], edge_port_fqdn_list: list[str], tt_number: str, process_id: UUIDstr, partner_name: str
) -> LSOState:
    """Perform a dry run of deploying :term:`BGP` peers."""
    extra_vars = {
        "subscription": subscription,
        "partner_name": partner_name,
        "verb": "deploy",
        "object": "bgp",
        "dry_run": True,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Deploying BGP peers for {subscription["description"]}",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/l3_core_service.yaml",
        "inventory": {"all": {"hosts": dict.fromkeys(edge_port_fqdn_list)}},
        "extra_vars": extra_vars,
    }


@step("[FOR REAL] Deploy BGP peers")
def deploy_bgp_peers_real(
    subscription: dict[str, Any], edge_port_fqdn_list: list[str], tt_number: str, process_id: UUIDstr, partner_name: str
) -> LSOState:
    """Deploy :term:`BGP` peers."""
    extra_vars = {
        "subscription": subscription,
        "partner_name": partner_name,
        "verb": "deploy",
        "object": "bgp",
        "dry_run": False,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Deploying BGP peers for {subscription["description"]}",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/l3_core_service.yaml",
        "inventory": {"all": {"hosts": dict.fromkeys(edge_port_fqdn_list)}},
        "extra_vars": extra_vars,
    }


@step("Check BGP peers")
def check_bgp_peers(subscription: dict[str, Any], edge_port_fqdn_list: list[str]) -> LSOState:
    """Check correct deployment of :term:`BGP` peers."""
    extra_vars = {"subscription": subscription, "verb": "check", "object": "bgp"}

    return {
        "playbook_name": "gap_ansible/playbooks/l3_core_service.yaml",
        "inventory": {"all": {"hosts": dict.fromkeys(edge_port_fqdn_list)}},
        "extra_vars": extra_vars,
    }


@step("Update Infoblox")
def update_dns_records(subscription: L3CoreService) -> State:
    """Update :term:`DNS` records in Infoblox."""
    #  TODO: implement
    return {"subscription": subscription}


@workflow(
    "Create L3 Core Service",
    initial_input_form=wrap_create_initial_input_form(initial_input_form_generator),
    target=Target.CREATE,
)
def create_l3_core_service() -> StepList:
    """Create a new L3 Core Service subscription including GÉANT IP and IAS.

    * Create subscription object in the service database
    * Deploy service binding ports
    * Deploy :term:`BGP` peers
    * Update :term:`DNS` records
    * Set the subscription in a provisioning state in the database
    """
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> lso_interaction(provision_sbp_dry)
        >> lso_interaction(provision_sbp_real)
        >> lso_interaction(check_sbp_functionality)
        >> lso_interaction(deploy_bgp_peers_dry)
        >> lso_interaction(deploy_bgp_peers_real)
        >> lso_interaction(check_bgp_peers)
        >> update_dns_records
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
